@@ -3,6 +3,7 @@ tn3270.emulator
 ~~~~~~~~~~~~~~~
 """
 
+from enum import Enum
 from itertools import chain
 import struct
 import logging
@@ -11,12 +12,18 @@ from .datastream import Command, WCC, Order, AID, parse_outbound_message, \
                         format_inbound_read_buffer_message, \
                         format_inbound_read_modified_message, \
                         format_inbound_structured_fields
-from .attributes import Attribute, HighlightExtendedAttribute, \
-                        ForegroundColorExtendedAttribute
+from .attributes import Attribute, AllExtendedAttribute, \
+                        Highlight, HighlightExtendedAttribute, \
+                        Color, ForegroundColorExtendedAttribute
 from .structured_fields import StructuredField, ReadPartitionType, \
                                QueryListRequestType, QueryCode
 from .ebcdic import DUP, FM
 
+class CharacterSet(Enum):
+    """Display cell character set."""
+
+    GE = 1
+
 class CellFormatting:
     """Display cell formatting."""
 
@@ -37,7 +44,12 @@ class CellFormatting:
                 self._apply_extended_attribute(extended_attribute)
 
     def _apply_extended_attribute(self, extended_attribute):
-        if isinstance(extended_attribute, HighlightExtendedAttribute):
+        if isinstance(extended_attribute, AllExtendedAttribute):
+            self.blink = False
+            self.reverse = False
+            self.underscore = False
+            self.color = 0x00
+        elif isinstance(extended_attribute, HighlightExtendedAttribute):
             self.blink = extended_attribute.blink
             self.reverse = extended_attribute.reverse
             self.underscore = extended_attribute.underscore
@@ -54,7 +66,7 @@ class CellFormatting:
 class Cell:
     """A display cell."""
 
-    def __init__(self, formatting=None):
+    def __init__(self, formatting):
         self.formatting = formatting
 
 class AttributeCell(Cell):
@@ -68,10 +80,11 @@ class AttributeCell(Cell):
 class CharacterCell(Cell):
     """A character display cell."""
 
-    def __init__(self, byte, formatting=None):
+    def __init__(self, byte, character_set=None, formatting=None):
         super().__init__(formatting)
 
         self.byte = byte
+        self.character_set = character_set
 
 class OperatorError(Exception):
     """Operator error."""
@@ -85,16 +98,34 @@ class FieldOverflowOperatorError(OperatorError):
 class Emulator:
     """TN3270 emulator."""
 
-    def __init__(self, stream, rows, columns):
+    def __init__(self, stream, rows, columns, supported_colors=8,
+            supported_highlights=[Highlight.BLINK, Highlight.REVERSE,
+                Highlight.UNDERSCORE]):
         self.logger = logging.getLogger(__name__)
 
         # TODO: Validate that stream has read_multiple() and write() methods.
         self.stream = stream
-        self.rows = rows
-        self.columns = columns
 
-        self.cells = [CharacterCell(0x00) for index in range(self.rows * self.columns)]
-        self.dirty = set(range(self.rows * self.columns))
+        if rows < 24:
+            raise ValueError('Invalid rows, must be at least 24')
+
+        if columns < 80:
+            raise ValueError('Invalid columns, must be at least 80')
+
+        if supported_colors not in [1, 4, 8]:
+            raise ValueError('Invalid suported colors, must be 1, 4 or 8')
+
+        self.default_dimensions = (24, 80)
+        self.alternate_dimensions = (rows, columns)
+
+        self.supported_colors = supported_colors
+        self.supported_highlights = supported_highlights
+
+        (self.rows, self.columns) = self.default_dimensions
+        self.alternate = False
+
+        self.cells = [CharacterCell(0x00) for index in range(rows * columns)]
+        self.dirty = set(range(rows * columns))
 
         self.address = 0
         self.cursor_address = 0
@@ -104,12 +135,18 @@ class Emulator:
 
     def update(self, **kwargs):
         """Read and execute outbound messages."""
+        if self.logger.isEnabledFor(logging.DEBUG):
+            self.logger.debug('Update')
+
         records = self.stream.read_multiple(**kwargs)
 
         if not records:
             return False
 
         for bytes_ in records:
+            if self.logger.isEnabledFor(logging.DEBUG):
+                self.logger.debug(f'\tRecord = {bytes_}')
+
             (command, *options) = parse_outbound_message(bytes_)
 
             self._execute(command, *options)
@@ -152,6 +189,10 @@ class Emulator:
 
     def home(self):
         """Home key."""
+        if not self.is_formatted():
+            self.cursor_address = 0
+            return
+
         addresses = self._get_addresses(0, (self.rows * self.columns) - 1)
 
         address = next((address for address in addresses
@@ -220,7 +261,8 @@ class Emulator:
 
         self._shift_left(self._wrap_address(self.cursor_address - 1), end_address)
 
-        attribute.modified = True
+        if attribute is not None:
+            attribute.modified = True
 
         self.cursor_address = self._wrap_address(self.cursor_address - 1)
 
@@ -229,29 +271,39 @@ class Emulator:
         if isinstance(self.cells[self.cursor_address], AttributeCell):
             raise ProtectedCellOperatorError
 
+        if not self.is_formatted():
+            raise NotImplementedError('Delete on unformatted screen is not supported')
+
         (_, end_address, attribute) = self.get_field(self.cursor_address)
 
         self._shift_left(self.cursor_address, end_address)
 
-        attribute.modified = True
+        if attribute is not None:
+            attribute.modified = True
 
     def erase_end_of_field(self):
         """Erase end of field (EOF) key."""
         if isinstance(self.cells[self.cursor_address], AttributeCell):
             raise ProtectedCellOperatorError
 
+        if not self.is_formatted():
+            raise NotImplementedError('Erase end of field on unformatted screen is not supported')
+
         (_, end_address, attribute) = self.get_field(self.cursor_address)
 
         for address in self._get_addresses(self.cursor_address, end_address):
-            self._write_character(address, 0x00, preserve_formatting=True)
+            self._write_character(address, 0x00, preserve=True)
 
         attribute.modified = True
 
     def erase_input(self):
         """Erase input key."""
+        if not self.is_formatted():
+            raise NotImplementedError('Erase input on unformatted screen is not supported')
+
         for (start_address, end_address, attribute) in self.get_fields():
             for address in self._get_addresses(start_address, end_address):
-                self._write_character(address, 0x00, preserve_formatting=True)
+                self._write_character(address, 0x00, preserve=True)
 
             attribute.modified = False
 
@@ -266,6 +318,10 @@ class Emulator:
 
         return bytes([self.cells[address].byte if isinstance(self.cells[address], CharacterCell) else 0x00 for address in addresses])
 
+    def is_formatted(self):
+        """Is there at least one attribute?"""
+        return any([isinstance(cell, AttributeCell) for cell in self.cells])
+
     def get_field(self, address):
         """Get the unprotected field containing or starting at the address."""
         (attribute, start_attribute_address) = self.find_attribute(address)
@@ -321,7 +377,6 @@ class Emulator:
     def _execute(self, command, *options):
         if self.logger.isEnabledFor(logging.DEBUG):
             self.logger.debug('Execute')
-            self.logger.debug(f'\tData    = {bytes_}')
             self.logger.debug(f'\tCommand = {command}')
 
         if command == Command.W:
@@ -331,7 +386,7 @@ class Emulator:
         elif command == Command.NOP:
             pass
         elif command in [Command.EW, Command.EWA]:
-            self._erase()
+            self._erase(command == Command.EWA)
             self._write(*options)
         elif command == Command.RM:
             self._read_modified()
@@ -342,17 +397,25 @@ class Emulator:
         elif command == Command.WSF:
             self._write_structured_fields(*options)
 
-    def _erase(self):
+    def _erase(self, alternate):
         self.logger.debug('Erase')
+        self.logger.debug(f'\tAlternate = {alternate}')
 
         self._clear()
 
+        if alternate:
+            (self.rows, self.columns) = self.alternate_dimensions
+        else:
+            (self.rows, self.columns) = self.default_dimensions
+
+        self.alternate = alternate
+
     def _erase_all_unprotected(self):
         self.logger.debug('Erase All Unprotected')
 
         for (start_address, end_address, attribute) in self.get_fields():
             for address in self._get_addresses(start_address, end_address):
-                self._write_character(address, 0x00, preserve_formatting=True)
+                self._write_character(address, 0x00, preserve=True)
 
             attribute.modified = False
 
@@ -372,7 +435,9 @@ class Emulator:
                 if isinstance(cell, AttributeCell):
                     cell.attribute.modified = False
 
-        formatting = None
+        field_formatting = None
+        character_formatting = None
+        character_set = None
 
         for (order, data) in orders:
             if self.logger.isEnabledFor(logging.DEBUG):
@@ -383,12 +448,15 @@ class Emulator:
                     self.logger.debug(f'\t\tParameters = {data}')
 
             if order == Order.PT:
-                # TODO: PT is more complex to implement that simply duplicating the
-                # behavior of tab() - how it behaves differs based on what command it
-                # follows.
-                raise NotImplementedError('PT order is not supported')
+                # TODO: Implement additional PT cases
+                if isinstance(self.cells[self.address], AttributeCell) and not self.cells[self.address].attribute.protected:
+                    self.address = self._wrap_address(self.address + 1)
+                else:
+                    raise NotImplementedError('PT order is not fully supported')
             elif order == Order.GE:
-                raise NotImplementedError('GE order is not supported')
+                self._write_character(self.address, data[0], CharacterSet.GE, None)
+
+                self.address = self._wrap_address(self.address + 1)
             elif order == Order.SBA:
                 if data[0] >= (self.rows * self.columns):
                     self.logger.warning(f'Address {data[0]} is out of range')
@@ -404,30 +472,30 @@ class Emulator:
                 unprotected_addresses = self._get_unprotected_addresses()
 
                 for address in unprotected_addresses.intersection(addresses):
-                    self._write_character(address, 0x00, preserve_formatting=True)
+                    self._write_character(address, 0x00, preserve=True)
 
                 self.address = stop_address
             elif order == Order.IC:
                 self.cursor_address = self.address
             elif order == Order.SF:
-                formatting = None
+                field_formatting = None
 
-                self._write_attribute(self.address, data[0], formatting)
+                self._write_attribute(self.address, data[0], field_formatting)
 
                 self.address = self._wrap_address(self.address + 1)
             elif order == Order.SA:
-                formatting = CellFormatting(formatting, extended_attributes=[data[0]])
+                character_formatting = CellFormatting(character_formatting, extended_attributes=[data[0]])
             elif order == Order.SFE:
                 # TODO: Confirm that formatting should be reset here
-                formatting = CellFormatting(None, extended_attributes=data[1])
+                field_formatting = CellFormatting(None, extended_attributes=data[1])
 
-                self._write_attribute(self.address, data[0] or Attribute(0), formatting)
+                self._write_attribute(self.address, data[0] or Attribute(0), field_formatting)
 
                 self.address = self._wrap_address(self.address + 1)
             elif order == Order.MF:
                 raise NotImplementedError('MF order is not supported')
             elif order == Order.RA:
-                (stop_address, byte) = data
+                (stop_address, byte, is_ge) = data
 
                 # TODO: Validate stop_address is in range...
                 end_address = self._wrap_address(stop_address - 1)
@@ -435,12 +503,12 @@ class Emulator:
                 addresses = self._get_addresses(self.address, end_address)
 
                 for address in addresses:
-                    self._write_character(address, byte, formatting)
+                    self._write_character(address, byte, CharacterSet.GE if is_ge else character_set, character_formatting)
 
                 self.address = stop_address
             elif order is None:
                 for byte in data:
-                    self._write_character(self.address, byte, formatting)
+                    self._write_character(self.address, byte, character_set, character_formatting)
 
                     self.address = self._wrap_address(self.address + 1)
 
@@ -453,33 +521,18 @@ class Emulator:
 
     def _clear(self):
         for address in range(self.rows * self.columns):
-            self._write_character(address, 0x00, None)
+            self._write_character(address, 0x00, None, None)
 
         self.address = 0
         self.cursor_address = 0
 
     def _read_buffer(self):
-        orders = []
-
-        data = bytearray()
-
-        for cell in self.cells:
-            if isinstance(cell, AttributeCell):
-                if data:
-                    orders.append((None, data))
-
-                    data = bytearray()
-
-                orders.append((Order.SF, [cell.attribute]))
-            elif isinstance(cell, CharacterCell):
-                data.append(cell.byte)
-
-        if data:
-            orders.append((None, data))
+        orders = self._generate_inbound_orders(0, (self.rows * self.columns) - 1)
 
         if self.logger.isEnabledFor(logging.DEBUG):
             self.logger.debug('Read Buffer')
             self.logger.debug(f'\tAID    = {self.current_aid}')
+            self.logger.debug(f'\tOrders = {orders}')
 
         bytes_ = format_inbound_read_buffer_message(self.current_aid, self.cursor_address, orders)
 
@@ -489,17 +542,25 @@ class Emulator:
         self.stream.write(bytes_)
 
     def _read_modified(self, all_=False):
-        modified_field_ranges = [(start_address, end_address) for (start_address, end_address, attribute) in self.get_fields() if attribute.modified]
+        if self.is_formatted():
+            modified_field_ranges = [(start_address, end_address) for (start_address, end_address, attribute) in self.get_fields() if attribute.modified]
+        else:
+            modified_field_ranges = [(0, (self.rows * self.columns) - 1)]
 
-        fields = [(start_address, self.get_bytes(start_address, end_address)) for (start_address, end_address) in modified_field_ranges]
+        orders = []
+
+        for (start_address, end_address) in modified_field_ranges:
+            orders.append((Order.SBA, [start_address]))
+
+            orders.extend(self._generate_inbound_orders(start_address, end_address, filter_null=True))
 
         if self.logger.isEnabledFor(logging.DEBUG):
             self.logger.debug('Read Modified')
             self.logger.debug(f'\tAID    = {self.current_aid}')
-            self.logger.debug(f'\tFields = {fields}')
+            self.logger.debug(f'\tOrders = {orders}')
             self.logger.debug(f'\tAll    = {all_}')
 
-        bytes_ = format_inbound_read_modified_message(self.current_aid, self.cursor_address, fields, all_)
+        bytes_ = format_inbound_read_modified_message(self.current_aid, self.cursor_address, orders, all_)
 
         if self.logger.isEnabledFor(logging.DEBUG):
             self.logger.debug(f'\tData   = {bytes_}')
@@ -507,30 +568,34 @@ class Emulator:
         self.stream.write(bytes_)
 
     def _input(self, byte, insert=False, move=True):
-        if isinstance(self.cells[self.cursor_address], AttributeCell):
-            raise ProtectedCellOperatorError
+        attribute = None
 
-        (_, end_address, attribute) = self.get_field(self.cursor_address)
+        if self.is_formatted():
+            if isinstance(self.cells[self.cursor_address], AttributeCell):
+                raise ProtectedCellOperatorError
 
-        if attribute is None or attribute.protected:
-            raise ProtectedCellOperatorError
+            (_, end_address, attribute) = self.get_field(self.cursor_address)
 
-        # TODO: Implement numeric field validation.
+            # TODO: Implement numeric field validation.
 
-        if insert and self.cells[self.cursor_address].byte != 0x00:
-            addresses = self._get_addresses(self.cursor_address, end_address)
+            if insert and self.cells[self.cursor_address].byte != 0x00:
+                addresses = self._get_addresses(self.cursor_address, end_address)
 
-            first_null_address = next((address for address in addresses
-                                       if self.cells[address].byte == 0x00), None)
+                first_null_address = next((address for address in addresses
+                                           if self.cells[address].byte == 0x00), None)
 
-            if first_null_address is None:
-                raise FieldOverflowOperatorError
+                if first_null_address is None:
+                    raise FieldOverflowOperatorError
 
-            self._shift_right(self.cursor_address, first_null_address)
+                self._shift_right(self.cursor_address, first_null_address)
+        else:
+            if insert:
+                raise NotImplementedError('Insert input on unformatted screen is not supported')
 
-        self._write_character(self.cursor_address, byte, preserve_formatting=True)
+        self._write_character(self.cursor_address, byte, preserve=True)
 
-        attribute.modified = True
+        if attribute is not None:
+            attribute.modified = True
 
         if not move:
             return
@@ -601,10 +666,10 @@ class Emulator:
 
         return self._wrap_address(address + 1)
 
-    def _write_attribute(self, index, attribute, formatting=None, preserve_formatting=False):
+    def _write_attribute(self, index, attribute, formatting=None, preserve=False):
         cell = self.cells[index]
 
-        if preserve_formatting:
+        if preserve:
             formatting = cell.formatting
 
         if isinstance(cell, AttributeCell):
@@ -620,20 +685,22 @@ class Emulator:
 
         return True
 
-    def _write_character(self, index, byte, formatting=None, preserve_formatting=False):
+    def _write_character(self, index, byte, character_set=None, formatting=None, preserve=False):
         cell = self.cells[index]
 
-        if preserve_formatting:
+        if preserve:
+            character_set = cell.character_set
             formatting = cell.formatting
 
         if isinstance(cell, CharacterCell):
-            if cell.byte == byte and cell.formatting == formatting:
+            if cell.byte == byte and cell.character_set == character_set and cell.formatting == formatting:
                 return False
 
             cell.byte = byte
+            cell.character_set = character_set
             cell.formatting = formatting
         else:
-            self.cells[index] = CharacterCell(byte, formatting)
+            self.cells[index] = CharacterCell(byte, character_set, formatting)
 
         self.dirty.add(index)
 
@@ -643,19 +710,50 @@ class Emulator:
         addresses = list(self._get_addresses(start_address, end_address))
 
         for (left_address, right_address) in zip(addresses, addresses[1:]):
-            self._write_character(left_address, self.cells[right_address].byte,
-                                  preserve_formatting=True)
+            self._write_character(left_address, self.cells[right_address].byte, preserve=True)
 
-        self._write_character(end_address, 0x00, preserve_formatting=True)
+        self._write_character(end_address, 0x00, preserve=True)
 
     def _shift_right(self, start_address, end_address):
         addresses = list(self._get_addresses(start_address, end_address))
 
         for (left_address, right_address) in reversed(list(zip(addresses, addresses[1:]))):
-            self._write_character(right_address, self.cells[left_address].byte,
-                                  preserve_formatting=True)
+            self._write_character(right_address, self.cells[left_address].byte, preserve=True)
 
-        self._write_character(start_address, 0x00, preserve_formatting=True)
+        self._write_character(start_address, 0x00, preserve=True)
+
+    def _generate_inbound_orders(self, start_address, end_address, filter_null=False):
+        orders = []
+
+        data = bytearray()
+
+        def eject():
+            nonlocal data
+
+            if data:
+                orders.append((None, data))
+
+                data = bytearray()
+
+        for address in self._get_addresses(start_address, end_address):
+            cell = self.cells[address]
+
+            if isinstance(cell, AttributeCell):
+                eject()
+
+                orders.append((Order.SF, [cell.attribute]))
+            elif isinstance(cell, CharacterCell):
+                if not filter_null or cell.byte != 0x00:
+                    if cell.character_set == CharacterSet.GE:
+                        eject()
+
+                        orders.append((Order.GE, [cell.byte]))
+                    else:
+                        data.append(cell.byte)
+
+        eject()
+
+        return orders
 
     def _write_structured_fields(self, structured_fields):
         if self.logger.isEnabledFor(logging.DEBUG):
@@ -712,7 +810,7 @@ class Emulator:
         if command == 0xf1:
             self._write(WCC(data[2]), data[3:])
         elif command in [0xf5, 0x7e]:
-            self._erase()
+            self._erase(command == 0x7e)
             self._write(WCC(data[2]), data[3:])
         elif command == 0x6f:
             self._erase_all_unprotected()
@@ -727,6 +825,7 @@ class Emulator:
 
         codes = [
             QueryCode.USABLE_AREA, QueryCode.ALPHANUMERIC_PARTITIONS,
+            QueryCode.COLOR, QueryCode.HIGHLIGHT,
             QueryCode.REPLY_MODES, QueryCode.IMPLICIT_PARTITIONS
         ]
 
@@ -743,13 +842,17 @@ class Emulator:
             reply = None
 
             if code == QueryCode.USABLE_AREA:
-                reply = _query_usable_area(self.rows, self.columns)
+                reply = _query_usable_area(self.alternate_dimensions)
             elif code == QueryCode.ALPHANUMERIC_PARTITIONS:
-                reply = _query_alphanumeric_partitions(self.rows, self.columns)
+                reply = _query_alphanumeric_partitions(self.alternate_dimensions)
+            elif code == QueryCode.COLOR:
+                reply = _query_color(self.supported_colors)
+            elif code == QueryCode.HIGHLIGHT:
+                reply = _query_highlight(self.supported_highlights)
             elif code == QueryCode.REPLY_MODES:
                 reply = _query_reply_modes()
             elif code == QueryCode.IMPLICIT_PARTITIONS:
-                reply = _query_implicit_partitions(self.rows, self.columns)
+                reply = _query_implicit_partitions(self.alternate_dimensions)
 
             if reply is not None:
                 replies.append((code, reply))
@@ -771,7 +874,9 @@ class Emulator:
 
         self.stream.write(bytes_)
 
-def _query_usable_area(rows, columns):
+def _query_usable_area(dimensions):
+    (rows, columns) = dimensions
+
     return struct.pack('>BBHHBHHHHBBH',
         0x01,           # 12/14-bit addressing allowed
         0x00,           # Cell units, no special features
@@ -787,19 +892,53 @@ def _query_usable_area(rows, columns):
         rows * columns  # Buffer size
     )
 
-def _query_alphanumeric_partitions(rows, columns):
+def _query_alphanumeric_partitions(dimensions):
+    (rows, columns) = dimensions
+
     return struct.pack('>BHB',
         1,              # One partition
         rows * columns, # Buffer size
         0x00            # No special features
     )
 
-def _query_reply_modes():
-    return struct.pack('>B',
-        0x00            # Field mode
+def _query_color(supported_colors):
+    reply = struct.pack('>BBBB',
+        0x00,           # Flags
+        8,              # Colors
+        0x00,           # Default
+        Color.GREEN
     )
 
-def _query_implicit_partitions(rows, columns):
+    for attribute in range(0xf1, 0xf7 + 1):
+        reply += struct.pack('>BB', attribute, attribute if supported_colors > 4 else 0x00)
+
+    return reply
+
+def _query_highlight(supported_highlights):
+    count = len(supported_highlights) + 1
+
+    reply = struct.pack('>BBB',
+        count,          # Types
+        0x00,           # Default
+        Highlight.NORMAL
+    )
+
+    for attribute in supported_highlights:
+        reply += struct.pack('>BB', attribute, attribute)
+
+    return reply
+
+def _query_reply_modes():
+    # TODO: Extended field mode is not really supported, but this is enough to trick z/VM
+    # into sending us pretty screens.
+    return struct.pack('>BB',
+        0x00,           # Field mode
+        0x01            # Extended field mode
+    )
+
+def _query_implicit_partitions(dimensions):
+    (rows, columns) = dimensions
+
     return struct.pack('>HBBBHHHH',
         0x0000,         # Flags
         0x0b,           # Length
